/*------------------------------------------------------------------------------

  Copyright (c) 2012-2014 viorica patraucean (vpatrauc@gmail.com)
  Copyright (c) 2018-2022 boshra rajaei      (bo.rajaei@gmail.com)

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.


  elsdc.c - This file belongs to ELSDc project (Ellipse and Line Segment
            Detector with continuous validation).
          - It contains the main detection procedure: curve growing and curve
            validation.

------------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
//#include <omp.h>
#include "misc.h"
#include "image.h"
#include "gauss.h"
#include "rectangle.h"
#include "iterator.h"
#include "polygon.h"
#include "ring.h"
#include "ellipse_fit.h"
#include "curve_grow.h"
#include "elsdc.h"


/*----------------------------------------------------------------------------*/
/** Continuous formulation of the number of false alarms
    NFAC = NT * k^n / n!
    log(n!) is bounded by Stirling's approximation:
       n! >= sqrt(2pi) * n^(n+0.5) * exp(-n)
     then, log10(NFA) <= log10(NT) + n*log10(k) - log10(latter expansion)
 */
#define LOG10_1_2_M_2__PI  0.39908993418
#define LOG10_EXP1         0.43429448190
static double NFAc( int n, double k, double logNT )
{
  /* check parameters */
  if( n < 0) error("NFAc: 'n' must be strictly positive.");
  if( n==0 ) return -logNT;
  double logNFAC = logNT;

  k = max(k,0.01);
  logNFAC = logNT + n * log10(k) - LOG10_1_2_M_2__PI -
            (n+0.5) * log10(n) + n * LOG10_EXP1 ;

  return -logNFAC;
}


/*----------------------------------------------------------------------------*/
/*----------------- FUNCTIONS FOR RECTANGLE VALIDATION -----------------------*/
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/** Compute additive normalised per-pixel error inside a rectangle, and stack
    visited points so they can be marked as 'used' if the rectangle is the most
    meaningful primitive.
 */
static void rect_count_error( PImageDouble angles, PImageInt used, PImageDouble gradmag,
                              Rectangle rec, double *err,
                              Point *in_buff, int size_in_buff, int *npts)//, int *npts1, int *npts2, int flag)
{

  int i, adr, xx, yy;
  double theta;

  /* check parameters */
  if( (angles == NULL) || (angles->data == NULL) )
    error("rect_count_error: invalid input 'angles' image.");
  if( (used == NULL) || (used->data == NULL) )
    error("rect_count_error: invalid input 'used' image.");
  if( in_buff == NULL ) error("rect_count_error: invalid buffer.");
  if( size_in_buff == 0 )
    error("rect_count_error: size_in_buff must be non null.");
  //if( out_buff == NULL ) error("rect_count_error: invalid buffer.");
  //if( size_out_buff == NULL )
  //  error("rect_count_error: size_in_buff must be non null.");

  theta = rec.theta - M_1_2_PI;
  *err = 0.0; //int flag1;
  //*size_out_buff = 0;


  for(i=0;i<size_in_buff;i++)
  {
      xx = in_buff[i].x;
      yy = in_buff[i].y;
      adr = xx + yy * used->xsize; //flag1 = 0;


      if( used->data[adr] != USED && angles->data[adr]!=NOTDEF)
      {
        //if(flag && i==size_in_buff-1)
        //        flag1 = 1;

        if(is_in_rect(rec, xx, yy))// && gradmag->data[adr]>min_grad)
        {

         (*err) += norm_angle_diff( angles->data[adr], theta );//gradmag->data[adr]*
         ++(*npts);//++(*npts1);
        }
        else
            {

         (*err) += +0.5+norm_angle_diff( angles->data[adr], theta );//gradmag->data[adr]*
         ++(*npts);

        }
      }
  }
}


/*----------------------------------------------------------------------------*/
/** Given a list of points, estimate segment. One segment is a particular case
    of a polygon.
 */
static void get_seg( PImageDouble angles, PImageDouble gradmag,
                     Rectangle *rec, Point *reg,
                     int reg_size, double prec )
{
  double sumdx = 0.0, sumdy = 0.0;
  double reg_angle;
  int i;
  int adr;

  /* check parameters */
  if( (angles == NULL) || (angles->data == NULL) )
    error("get_seg: invalid input 'angles' image.");
  if( (gradmag == NULL) || (gradmag->data == NULL) )
    error("get_seg: invalid input 'angles' image.");
  //if( (used == NULL) || (used->data == NULL) )
  //  error("get_seg: invalid input 'used' image.");
  if( rec == NULL ) error("get_seg: invalid input polygon.");
  if( reg == NULL ) error("get_seg: invalid input region.");
  if( reg_size <= 0 )
    error("get_seg: 'reg_size' must be strictly positive.");

  /* compute the mean angle of the whole region */
  for( i=0; i<reg_size; i++ )
    {
      adr = reg[i].x + reg[i].y * angles->xsize;
      sumdx += cos( angles->data[adr] );
      sumdy += sin( angles->data[adr] );
    }
  reg_angle = atan2( sumdy, sumdx );

  /* estimate rectangle */
  region2rect( gradmag, reg, 0, reg_size, reg_angle, rec, prec, 0 );

}

/*----------------------------------------------------------------------------*/
/** Compute additive normalised per-pixel error inside a circular ring, and
    stack visited points so they can be marked as 'used' if the circle ring is
    the most meaningful primitive.
 */
static void circ_count_error( PImageDouble angles, PImageInt used, PImageDouble gradmag, Ring cring,
                              Point *in_buff, int size_in_buff,
                              int grad_dir, int *npts, double *err)//, double min_grad )
{
  int xx, yy, i, adr;
  double theta;

  /* check parameters */
  if( (angles == NULL) || (angles->data == NULL) )
    error("circ_count_error: invalid input 'angles' image.");
  if( (used == NULL) || (used->data == NULL) )
    error("circ_count_error: invalid input 'used' image.");
  if( in_buff == NULL ) error("circ_count_error: invalid local buffer.");
  //if( out_buff == NULL ) error("circ_count_error: invalid local buffer.");
  if( size_in_buff == 0 )
    error("circ_count_error: 'size_local_buff' must be non zero.");
  //if( *label <= 1 ) error("circ_count_error: forbidden label value.");


  *err = 0.0;
  //*size_out_buff = 0;


for(i=0;i<size_in_buff;i++)
  {
      xx = in_buff[i].x;
      yy = in_buff[i].y;
      adr = xx + yy * used->xsize;

      if( used->data[adr] != USED && angles->data[adr]!=NOTDEF)
      {
        if(is_in_circ_ring( &cring, xx, yy ))// && gradmag->data[adr]>min_grad )
        {
         theta = atan2( (double)yy - cring.cy,
                        (double)xx - cring.cx );
         if( grad_dir == 0 )
         {
		    if( theta > 0 ) theta = -(M_PI-theta);
		     else theta = M_PI + theta;
	     }
         (*err) += norm_angle_diff( angles->data[adr], theta );//gradmag->data[adr]*
         ++(*npts);
        }
        else
        {

         theta = atan2( (double)yy - cring.cy,
                        (double)xx - cring.cx );
         if( grad_dir == 0 )
         {
		    if( theta > 0 ) theta = -(M_PI-theta);
		     else theta = M_PI + theta;
	     }
         (*err) += 0.5+norm_angle_diff( angles->data[adr], theta );//gradmag->data[adr]*
         ++(*npts);
        }
      }
  }

}

/*----------------------------------------------------------------------------*/
/** Compute additive normalised per-pixel error inside an elliptical ring, and
    stack visited points so they can be marked as 'used' if the ellipse ring is
    the most meaningful primitive.
 */
static void ell_count_error( PImageDouble angles, PImageInt used, PImageDouble gradmag, Ring ering,
                              double *foci, Point *in_buff, int size_in_buff,
                              int grad_dir, int *npts, double *err)//, double min_grad )
{
  int xx, yy, i, adr;
  double theta;

  /* check parameters */
  if( (angles == NULL) || (angles->data == NULL) )
    error("ell_count_error: invalid input 'angles' image.");
  if( (used == NULL) || (used->data == NULL) )
    error("ell_count_error: invalid input 'used' image.");
  if( in_buff == NULL ) error("ell_count_error: invalid local buffer.");
  //if( out_buff == NULL ) error("circ_count_error: invalid local buffer.");
  if( size_in_buff == 0 )
    error("ell_count_error: 'size_local_buff' must be non zero.");
  //if( *label <= 1 ) error("ell_count_error: forbidden label value.");


  *err = 0.0;
  //*size_out_buff = 0;


for(i=0;i<size_in_buff;i++)
  {
      xx = in_buff[i].x;
      yy = in_buff[i].y;
      adr = xx + yy * used->xsize;

      if( used->data[adr] != USED && angles->data[adr]!=NOTDEF)
      {

        theta = ellipse_normal_angle( (double)xx, (double)yy, foci);
        if( grad_dir == 0 )
         {
		    if( theta > 0 ) theta = -(M_PI-theta);
		     else theta = M_PI + theta;
	     }

        if(is_in_ell_ring( &ering, xx, yy ))// && gradmag->data[adr]>min_grad )
        {
         (*err) += norm_angle_diff( angles->data[adr], theta );//gradmag->data[adr]*
         ++(*npts);
        }
        else
        {
         (*err) += 0.5+norm_angle_diff( angles->data[adr], theta );//gradmag->data[adr]*
         ++(*npts);

        }
      }

  }

}



/*----------------------------------------------------------------------------*/
/*------------------------------- BUILD OUTPUT -------------------------------*/
/*----------------------------------------------------------------------------*/

/** This code has to be removed from the final version
*/
void debug_output (Point *reg, int reg_size, int flag)
{
    FILE *txt_file;
    int i;

    if(!flag)
    {
        if( (txt_file = fopen("raw_reg.txt","w")) == NULL )
        error("main: can't open output raw file.");
    }
    else if(flag==1)
    {
        if( (txt_file = fopen("local_reg1.txt","w")) == NULL )
        error("main: can't open output local file.");

    }
    else
    {
        if( (txt_file = fopen("local_reg2.txt","a+")) == NULL )
        error("main: can't open output local file.");

    }

    //fprintf( txt_file,"%d ", reg_size );
      for( i=0; i<reg_size; i++ )
        {
          fprintf( txt_file,"%d %d \n",
                   reg[i].x, reg[i].y );
        }
      fclose(txt_file);
}


/*----------------------------------------------------------------------------*/
/** A meaningful polygon has been found; store it in output.
 */
void add_seg_out( Polygon **poly_out, Rectangle *rec, int poly_count,
                   int *poly_count_max, int **labels, int lbl )
{

  if( poly_count>*poly_count_max )
    {
      *poly_out = (Polygon *) realloc ( *poly_out, (2*poly_count) *
                                                   sizeof(Polygon) );

      *labels = (int *) realloc ( *labels, (2*poly_count) * sizeof(int) );
      *poly_count_max = 2 * poly_count;
    }

  //polyrect2polygon( poly, &( (*poly_out)[poly_count-1] ) );

  (*poly_out)[poly_count-1].dim = 2 ;
  (*poly_out)[poly_count-1].pts = (PointD *) malloc ( 2 * sizeof(PointD) );
  if( (*poly_out)[poly_count-1].pts == NULL )
    error("add_seg_out: not enough memory.");

  /* copy ending points */
  (*poly_out)[poly_count-1].pts[0].x = rec->x1 + 0.5;
  (*poly_out)[poly_count-1].pts[0].y = rec->y1 + 0.5;
  (*poly_out)[poly_count-1].pts[1].x = rec->x2 + 0.5;
  (*poly_out)[poly_count-1].pts[1].y = rec->y2 + 0.5;

  (*labels)[poly_count-1] = lbl;
}

/*----------------------------------------------------------------------------*/
/** store a dot in one cutting point along a meaningful chain in output.
 */
void add_dot_out( Dot **dot_out, double *center, int dot_count,
                   int *dot_count_max)
{

  if( dot_count>*dot_count_max )
    {
      *dot_out = (Dot *) realloc ( *dot_out, (2*dot_count) *
                                                   sizeof(Dot) );

      *dot_count_max = 2 * dot_count;
    }

  (*dot_out)[dot_count-1].cx = center[0];
  (*dot_out)[dot_count-1].cy = center[1];
  (*dot_out)[dot_count-1].rad = 2;
  (*dot_out)[dot_count-1].width = 2;
}

/*----------------------------------------------------------------------------*/
/** A meaningful ellipse has been found; store it in output.
 */
void add_ell_out( Ring **ell_out, Ring *ell, int ell_count, int *ell_count_max,
                  int **labels, int lbl )
{
  if( ell_count>*ell_count_max )
    {
      *ell_out = ( Ring *) realloc ( *ell_out, (2*ell_count) * sizeof(Ring) );
      *labels =  ( int *) realloc ( *labels, (2*ell_count) * sizeof(int) );
      *ell_count_max = 2 * ell_count;
    }

  copy_ring( ell, &( (*ell_out)[ell_count-1] ) );
  (*labels)[ell_count-1] = lbl;
}

/*----------------------------------------------------------------------------*/
/** Entry point in detector's code. Gets as input an image, and returns two
    lists of geometric shapes: polygons and ellipses (circles are considered
    particular ellipses, and segments are particular polygons).
 */
void ELSDc( PImageDouble in, double *in_param, int *ell_count, Ring **ell_out, int **ell_labels,
            int *poly_count, Polygon **poly_out, int **poly_labels,
            int *dot_count, Dot **dot_out, PImageInt out )
{
  double ang_th1 = in_param[1];     /* gradient angle tolerance in degrees for region growing */
  //double ang_th2 = in_param[2];     /* gradient angle tolerance in degrees for possible region merging */
  double prec1;              /* radian precisions */ //, prec2
  double p;
  double alpha = in_param[4];      /*in NFA validation step after a region has been detected, only points with gradient magnitude larger
                            than alpha times of the seed point are considered */
  double rho2 = in_param[3];     /*in NFA validation step after a region has been detected, only points with gradient magnitude larger
                            than this threshold are considered */
  double rho = in_param[2];         /* in the region growing step, pixels with gradient magnitude under this
                               threshold are not considered, as their (weak)
                               orientation might be too affected by noise  */
  //int grad_dirc;            /* direction of the gradient wrt center of the
  //                             circle */
  int grad_dire;            /* direction of the gradient wrt center of the
                               ellipse */
  double density_th = in_param[5];  /* minimum density */
  int max_cut_pnt = in_param[0];    /*To control time complexity, the maximum number of possible cut points along a chain of subregions */
  void *mem_p;              /* parameters used for sorting */
  int n_bins = 1024;        /* (binning) pixels according to */
  double max_grad = 255.0;  /* to their gradient magnitude */
  CoordList *list_p;        /* list of sorted pixels */
  unsigned int xsize,ysize; /* image size */
  double mlog10eps = 0.0;   /* minus log of the number of accepted false
                               positives ( eps=1 ) */
  /*int min_size_chain;        minimum size of a region that could contain a
                               meaningful chain */
  double msize;             /* max value allowed for circle/ellipse radius */

  double best_nfa; /* NFA values using the continuous */

  parList *best_par = NULL; /* best partitioning of current chain*/
  PImageDouble imgauss;     /* smooth version of the original image, used during
                               candidate selection */
  PImageDouble angles;      /* gradient angles smooth image */
  PImageDouble angles0;     /* gradient angles original image */
  PImageDouble gradmag;     /* gradient magnitude */
  PImageDouble gradx, grady;/* gradient orientations on Ox and Oy */
  PImageInt used;           /* image to mark already used points */
  PolyRect *poly;           /* parameters of polygon as list of rectangles */

  Ring ering;               /* ring parameters */

  Point *reg, *local_region; /* list of points gathered during curve growing */
  int reg_size, local_reg_size;

  Rectangle rec;

  //For memory allocation to structures of detecting ellipses, lines and dots
  int ell_count_max = 0;
  int poly_count_max = 0;
  int dot_count_max = 0;

  double wmax, wmin;

  int adr, j;
  //double cparam[5];
  double eparam[5];
  double pext1[2], pext2[2], ang0;
  int label = 2;            /* 0 is NOTUSED, 1 is USED, so 'label' must go
                               upward from 2 */
  int label_meaningful = 0;
  //double spir;
  double foci[4];

  parList *partition, *strpntr; /* list of possible partitions along a detected chain */
  //parList *par; /*a pointer to the current partition item being processed by one thread */
  int parcnt; //number of possible partitions along a detected chain

  /* check parameters */
  if( (in == NULL) || (in->data == NULL) ) error("ELSDc: invalid input image.");
  if( (out == NULL) || (out->data == NULL) )
    error("ELSDc: invalid output label image.");

  /* allocate and init variables */
  prec1 = M_PI*ang_th1/180.0;
  //prec2 = M_PI*ang_th2/180.0;
  p = ang_th1/180.0;
  xsize = in->xsize;
  ysize = in->ysize;

  /* allocate buffers to store point coordinates */
  reg = (Point *) malloc( xsize * ysize * sizeof(Point) );
  if( reg == NULL ) error("ELSDc: not enough memory.");
  //local_region = (Point *) malloc( xsize * ysize * sizeof(Point) );
  //if( local_region == NULL ) error("ELSDc: not enough memory.");
  //partion = (parList*) malloc( sizeof(parList) );
  //if( partion == NULL ) error("ELSDc: not enough memory.");

  used = new_PImageInt_ini( xsize, ysize, NOTUSED );


  /* perform gaussian smoothing */
  imgauss = gaussian_sampler( in, 1.0, 0.6 );

  /* compute gradient magnitude and orientation  for the smooth image */
  img_gradient_sort( imgauss, rho, &list_p, &mem_p, n_bins, max_grad, &angles,
                     &gradmag, &gradx, &grady );

  /* compute gradient orientation for the original image */
  angles0 = img_gradient_angle( in, rho );

  /* input image not needed any more; free it */
  free_PImageDouble(in);

  msize = max(xsize,ysize);

  poly = new_polyrect();

  double min_grad; //int cntfixed;


  /* begin primitive detection */
  for( ; list_p; list_p = list_p->next )
  {
      adr = list_p->y * used->xsize + list_p->x;
      /* only pixels that haven't been used already can be seed pixels */
      if( used->data[adr] == USED ) continue;

      /* only pixels with defined gradient can be seed pixels */
      if( angles->data[adr] == NOTDEF ) continue;

      if(gradmag->data[adr]<rho2)
        break;

      /* only pixels that are local maxima in the image can be seed pixels */
      /*if( !local_max( imgauss, list_p->x, list_p->y ) ) continue; */

      /* init intra-iteration variables */
      best_nfa = mlog10eps;

      label++;
      //spir = 0.0;
      clear_polyrect(poly);


      /* set initial seed point */
      reg[0].x = list_p->x;
      reg[0].y = list_p->y;
      reg_size = 1;

      /* Gather points aligned along a convex polygon */
      if( !curve_grow( gradmag, angles, used, reg, &reg_size, density_th,
                       prec1, poly, &label, max_cut_pnt+1) ) //, pext2, pext1, &spir
        continue;

      //printf("\n#segments: %d\n",poly->dim);

      min_grad = alpha*gradmag->data[adr];

      /* generate all possible ways of partitioning the detected chain */
      //cntfixed = 0;
      partition = (parList*) malloc( sizeof(parList) );
      if( partition == NULL ) error("ELSDc: not enough memory.");
      generate_partition(partition, &parcnt, *poly, max_cut_pnt);//, &cntfixed);
      strpntr = partition;


      #pragma omp parallel for schedule(dynamic) //private(flag,npts,ang0,grad_dire,curr_nfa)
      for(j=0;j<parcnt;j++){

      parList *par;

       #pragma omp critical
        {
            par = partition;
            partition = partition->next;
        }


        double *err = (double *) malloc( par->length * sizeof(double) );
        Point *local_reg = (Point *) malloc( reg_size * sizeof(Point) );

        double param[5];
        double pex1[2], pex2[2], foc[4], ang;
        Rectangle rect;
        Ring ring;
        int local_reg_si, min_size_chain, flag, npts, grad_dir;

        double logNT = par->length*LOG10_3 + par->length*LOG10_2 +
                (par->length+1+0.5)*(log10((double)xsize) + log10((double)ysize))+
                (1.5*sum_arr(par->prim_type,par->length,1))*(log10((double)xsize) + log10((double)ysize))+//ellipse
                (0.5*sum_arr(par->prim_type,par->length,2))*(log10((double)xsize) + log10((double)ysize));//circle
        min_size_chain = (int)( (-logNT - mlog10eps) / log10(p) );

         if(reg_size>min_size_chain && local_reg != NULL && err!= NULL){

        flag = 0; npts = 0;

         for(int i=0;i<par->length;i++)
         {
            if(par->prim_type[i]==0)//prim_type[i]=0 indicates line segment
            {


                make_local_region(reg, reg_size, local_reg, &local_reg_si, poly,
                          par->cut_pnt, i, par->length, pex2, pex1, gradmag, min_grad );

                if(local_reg_si<=2) {flag = 1;break; }


                    get_seg( angles, gradmag, &rect, local_reg, local_reg_si, prec1 );
                    rect.wmin = poly->wmin; rect.wmax = poly->wmax;

                    rect_count_error( angles0, used, gradmag, rect, &err[i], local_reg, local_reg_si,
                                 &npts);

             }
             else if(par->prim_type[i]==1){//prim_type[i]=1 indicates ellipse arc

                make_local_region(reg, reg_size, local_reg, &local_reg_si, poly,
                          par->cut_pnt, i, par->length, pex2, pex1, gradmag, min_grad );

                if(local_reg_si<=2){flag = 1;break; }


                    conic_fit( gradx, grady, local_reg, local_reg_si, param, 0);


                    ang = angles->data[ local_reg[0].y*angles->xsize + local_reg[0].x ];
                    if( get_ring( local_reg, local_reg_si, ang, param, pex1, pex2, ELLIPSE,
                       &ring, &grad_dir, foc, msize ) ) //spir,
                    {
                        ring.wmin = poly->wmin; ring.wmax = poly->wmax;

                        ell_count_error( angles0, used, gradmag, ring, foc, local_reg, local_reg_si,
                                    grad_dir, &npts, &err[i] );

                    }
                    else  {flag = 1;break; }


            }
            else{//prim_type[i]=2 indicates circle arc

                make_local_region(reg, reg_size, local_reg, &local_reg_si, poly,
                          par->cut_pnt, i, par->length, pex2, pex1, gradmag, min_grad );

                if(local_reg_si<=2){flag = 1;break; }

                    conic_fit( gradx, grady, local_reg, local_reg_si, param, 1);

                    ang = angles->data[ local_reg[0].y*angles->xsize + local_reg[0].x ];
                    if( get_ring( local_reg, local_reg_si, ang, param, pex1, pex2, CIRCLE,
                       &ring, &grad_dir, foc, msize ) )//spir,
                    {
                        ring.wmin = poly->wmin; ring.wmax = poly->wmax;

                        circ_count_error( angles0, used, gradmag, ring, local_reg, local_reg_si,
                                    grad_dir, &npts, &err[i]);

                    }
                    else  {flag = 1;break; }



            }

         }

         if(!flag) {

         double curr_nfa = NFAc(npts, sum_err(err,par->length), logNT);


         #pragma omp critical
         {
             if( curr_nfa > best_nfa ){
                best_nfa = curr_nfa;
                best_par = par;
            }

         }


         }


      }
      if(err!=NULL) free(err); if(local_reg!=NULL) free(local_reg);
      }

      if( best_nfa > mlog10eps ){
         label_meaningful++;
         //printf("\n#part:     %d\n\n",best_par->length);

         local_region = (Point *) malloc( reg_size * sizeof(Point) );
         if( local_region == NULL ) error("ELSDc: not enough memory.");
         //eparam = (double *) malloc( 5 * sizeof(double) );
         //foci = (double *) malloc( 4 * sizeof(double) );
         wmin = poly->wmin;
         wmax = poly->wmax;

         for(int i=0;i<best_par->length;i++)
         {


             if(best_par->prim_type[i]==0){

                  make_local_region(reg, reg_size, local_region, &local_reg_size, poly,
                           best_par->cut_pnt, i, best_par->length, pext2, pext1, gradmag, min_grad );

                  get_seg( angles, gradmag, &rec, local_region, local_reg_size, prec1 );

                  rec.wmin = wmin; rec.wmax = wmax;

                  (*poly_count)++;
                  add_seg_out( poly_out, &rec, *poly_count,
                            &poly_count_max, poly_labels, label_meaningful );
              }
             else if(best_par->prim_type[i]==1){

                  make_local_region(reg, reg_size, local_region, &local_reg_size, poly,
                             best_par->cut_pnt, i, best_par->length, pext2, pext1, gradmag, min_grad );

                  conic_fit( gradx, grady, local_region, local_reg_size, eparam, 0);

                  ang0 = angles->data[ local_region[0].y*angles->xsize + local_region[0].x ];
                  get_ring( local_region, local_reg_size, ang0, eparam, pext1, pext2, ELLIPSE ,
                       &ering, &grad_dire, foci, msize );//spir,

                  ering.wmax = wmax; ering.wmin = wmin;

                  (*ell_count)++;
                  add_ell_out(ell_out, &ering, *ell_count, &ell_count_max,
                                ell_labels, label_meaningful );

              }
              else{

                  make_local_region(reg, reg_size, local_region, &local_reg_size, poly,
                             best_par->cut_pnt, i, best_par->length, pext2, pext1, gradmag, min_grad );

                  conic_fit( gradx, grady, local_region, local_reg_size, eparam, 1);

                  ang0 = angles->data[ local_region[0].y*angles->xsize + local_region[0].x ];

                  get_ring( local_region, local_reg_size, ang0, eparam, pext1, pext2, CIRCLE,
                       &ering, &grad_dire, foci, msize );//spir,

                  ering.wmax = wmax; ering.wmin = wmin;

                  (*ell_count)++;
                  add_ell_out(ell_out, &ering, *ell_count, &ell_count_max,
                                ell_labels, label_meaningful );

              }


             if(best_par->length>1 && i<best_par->length-1){
                 (*dot_count)++;
                 add_dot_out(dot_out, pext2, *dot_count, &dot_count_max);
              }



         }
         /* Mark as USED the pixels of the best feature in 'used' */
         mark_img_pts( used, reg, 0, reg_size, USED );

         /* Mark label of the best feature in output label image */
         mark_img_pts( out, reg, 0, reg_size, label_meaningful );
         free(local_region);
      }
      clear_partition(strpntr);
      //getchar();

   }  /* ! for loop */

   printf("\nNumber of chains detected = %d\n", label_meaningful);

  /* free useless structures and arrays */
  free_PImageDouble(imgauss);
  free_PImageDouble(gradx);
  free_PImageDouble(grady);
  free_PImageDouble(gradmag);
  free_PImageDouble(angles);
  free_PImageDouble(angles0);
  free_PImageInt(used);
  free(reg);
  free(poly->rectlist); free(poly);
  free(mem_p);

}
